﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;              
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis; 
 
namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Regression {
  /// <summary>
  /// Analyzer to use in combination with eps-lexicase selection
  /// </summary>
  [Item("EpsLexicaseAnalyzer", "")]
  [StorableClass]
  public sealed class EpsLexicaseAnalyzer : SymbolicDataAnalysisSingleObjectiveAnalyzer {

    [StorableConstructor]
    private EpsLexicaseAnalyzer(bool deserializing) : base(deserializing) { }
    private EpsLexicaseAnalyzer(EpsLexicaseAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new EpsLexicaseAnalyzer(this, cloner);
    }

    public EpsLexicaseAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<IRegressionProblemData>("ProblemData"));
      Parameters.Add(new LookupParameter<ISymbolicDataAnalysisExpressionTreeInterpreter>("SymbolicExpressionTreeInterpreter"));
      Parameters.Add(new LookupParameter<DoubleLimit>("EstimationLimits"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleArray>("Errors", 1));
    }

    public override IOperation Apply() {
      var problemData = (IRegressionProblemData)Parameters["ProblemData"].ActualValue;
      var ds = problemData.Dataset;
      var trainingRows = problemData.TrainingIndices.ToArray();
      var y = problemData.TargetVariableTrainingValues.ToArray();
      var trees = SymbolicExpressionTree;
      var interpreter = (ISymbolicDataAnalysisExpressionTreeInterpreter)Parameters["SymbolicExpressionTreeInterpreter"].ActualValue;
      var limits = (DoubleLimit)Parameters["EstimationLimits"].ActualValue;

      var errors = new ItemArray<DoubleArray>(trees.Length);
      int i = 0;
      foreach (var tree in trees) {
        var model = new SymbolicRegressionModel(problemData.TargetVariable, tree, (ISymbolicDataAnalysisExpressionTreeInterpreter)interpreter.Clone(), limits.Lower, limits.Upper);
        if (ApplyLinearScalingParameter.ActualValue.Value)
          model.Scale(problemData);

        errors[i] = new DoubleArray(y.Zip(model.GetEstimatedValues(ds, trainingRows), (yi, pi) => Math.Abs(yi - pi)).ToArray());
        i++;
      }
      Parameters["Errors"].ActualValue = errors;

      return base.Apply();
    }
  }
}
